'''Python Abstraction for a Reaction.'''

from symmathSBML import constants as cn
from symmathSBML.kinetic_law import KineticLaw
from symmathSBML import util

import numpy as np
import libsbml


class Reaction(object):

  def __init__(self, libsbml_reaction, function_definitions=None):
    """
    :param libsbml.Reaction libsbml_reaction:
    :param function_definitions list-FunctionDefinition:
    """
    self.reaction = libsbml_reaction
    # List of species reference
    self.reactants =  [self.reaction.getReactant(n)
        for n in range(self.reaction.getNumReactants())]
    # List of species reference
    self.products =  [self.reaction.getProduct(n)
        for n in range(self.reaction.getNumProducts())]
    # BUGFIX: Used None in function_defintiions=
    self.kinetic_law = KineticLaw(
        self.reaction.getKineticLaw(), self,
        function_definitions=function_definitions)
    self.id = self.reaction.getId()

  def getId(self):
    return self.id

  def __repr__(self):
    reactant_str = " + ".join([r.getSpecies() for r in self.reactants])
    product_str = " + ".join([p.getSpecies() for p in self.products])
    try:
        if self.kinetic_law.expanded_formula is not None:
          kinetic_str = self.kinetic_law.expanded_formula
        else:
          kinetic_str = self.kinetic_law.formula
    except AttributeError:
        kinetic_str = "Kinetic law not calculated."
    return "%s -> %s; %s" % (reactant_str, product_str, kinetic_str)



'''Provides Information on SBML Kinetics Laws'''


from symmathSBML import constants as cn
from symmathSBML import util
from symmathSBML import exceptions
from symmathSBML import msgs

import collections
import numpy as np
import re

MAX_RECURSION = 5
# Tokens used in mathematical expressions of SBML function definitions
TOKENS = [" ", "(", ")", "/", "-", "*", "+", ","]


class KineticLaw(object):

  def __init__(self, libsbml_kinetics, reaction, function_definitions=None):
    """
    :param libsbml.KineticLaw libsbml_kinetics:
    :param function_definitions list-FunctionDefinition:
    """
    self.reaction = reaction
    # libsbml object for kinetics
    self.libsbml_kinetics = libsbml_kinetics
    # String version of chemical formula
    self.formula = self.libsbml_kinetics.getFormula()
    # Parameters and chemical species
    self.symbols = self._getSymbols()
    # Reaction for the kinetics law
    self.reaction = reaction
    # Expanded kinetic formula (remove embedded functions)
    if function_definitions is None:
      self.expanded_formula = None
    else:
      self.expandFormula(function_definitions)
    self.expression_formula = None  # valid symPy expression string

  def __repr__(self):
    return self.formula

  def expandFormula(self, function_definitions):
    """
    Expands the kinetics formula, replacing function definitions
    with their body.

    Parameters
    ----------
    function_definitions: list-FunctionDefinition
    """
    self.expanded_formula = self._expandFormula(
        self.formula, function_definitions)

  def mkSymbolExpression(self, function_definitions):
    """
    Creates a string that can be processed by sympy.
    
    Parameters
    -------
    function_definitions: list-FunctionDefinition
    
    Returns
    -------
    str
    """
    if self.expanded_formula is None:
      self.expandFormula(function_definitions)
    self.expression_formula = str(self.expanded_formula)
    self.expression_formual.replace('^','**')

  @staticmethod
  def _expandFormula(expansion, function_definitions,
          num_recursion=0):
      """
      Expands the kinetics formula, replacing function definitions
      with their body.
  
      Parameters
      ----------
      expansion: str
          expansion of the kinetic law
      function_definitions: list-FunctionDefinition
      num_recursion: int
      
      Returns
      -------
      str
      """
      if num_recursion > MAX_RECURSION:
          return expansion
      done = True
      for fd in function_definitions:
          # Find the function calls
          calls = re.findall(r'{}\(.*?\)'.format(fd.id), expansion)
          if len(calls) == 0:
            continue
          done = False
          for call in calls:
            # Find argument call. Ex: '(a, b)'
            call_arguments = re.findall(r'\(.*?\)', call)[0]
            call_arguments = call_arguments.strip()
            call_arguments = call_arguments[1:-1]  # Eliminate parentheses
            arguments = call_arguments.split(',')
            arguments = [a.strip() for a in arguments]
            body = str(fd.body)
            tokenized_body = util.tokenize(body, TOKENS)
            for pos, token in enumerate(tokenized_body):
                # Need a bit of sophistication with replacing arguments to
                # avoid problems with substrings
                for formal_arg, call_arg in zip(fd.argument_names, arguments):
                    if token == formal_arg:
                        tokenized_body[pos] = call_arg
                        # Don't replace a replaced string
                        break   
            new_body = util.detokenize(tokenized_body)
            expansion = expansion.replace(call, new_body)
      if not done:
          return KineticLaw._expandFormula(expansion, function_definitions,
              num_recursion=num_recursion+1)
      return expansion
 

  def _getSymbols(self):
    """
    Finds the parameters and species names for the
    kinetics law. Exposing this information requires
    a recursive search of the parse tree for the
    kinetics expression.
    :return list-str:
    """
    global cur_depth
    MAX_DEPTH = 20
    cur_depth = 0
    def augment(ast_node, result):
      global cur_depth
      cur_depth += 1
      if cur_depth > MAX_DEPTH:
        import pdb; pdb.set_trace()
        raise exceptions.BadKineticsMath(str(self.reaction))
      for idx in range(ast_node.getNumChildren()):
        child_node = ast_node.getChild(idx)
        if child_node.getName() is None:
          additions = augment(child_node, result)
          result.extend(additions)
        else:
          if child_node.isFunction():
            additions = augment(child_node, result)
            result.extend(additions)
          else:
            result.append(child_node.getName())
      cur_depth -= 1
      return result

    ast_node = self.libsbml_kinetics.getMath()
    if ast_node.getName() is None:
      result = []
    else:
      result = [ast_node.getName()]
    return augment(ast_node, result)



"""Constants for Project."""
import collections
import os


################ DIRECTORIES #################
PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
for _ in range(1):
  PROJECT_DIR = os.path.dirname(PROJECT_DIR)
CODE_DIR = os.path.join(PROJECT_DIR, "symmathSBML")
TEST_DIR = os.path.join(PROJECT_DIR, "tests")
DATA_DIR = os.path.join(PROJECT_DIR, "data")
BIOMODELS_ZIP_FILENAME = "biomodels.zip"

############### TEST FILES #####################
TEST_FILE = os.path.join(TEST_DIR, "test_file.xml")
TEST_FILE2 = os.path.join(TEST_DIR, "test_file2.xml")
# test_file3: originally curated_017
TEST_FILE3 = os.path.join(TEST_DIR, "test_file3.xml")
TEST_FILE5 = os.path.join(TEST_DIR, "test_file5.antimony")
TEST_FILE4 = os.path.join(TEST_DIR, "test_file4.xml")
# test_file6: originally curated_050
TEST_FILE6 = os.path.join(TEST_DIR, "test_file6.xml")
# test_file7: originally curated_018
TEST_FILE7 = os.path.join(TEST_DIR, "test_file7.xml")
# test_file8: originally curated_004
TEST_FILE8 = os.path.join(TEST_DIR, "test_file8.xml")
# test_file8: originally curated_008
TEST_FILE9 = os.path.join(TEST_DIR, "test_file9.xml")
# test_file10: originally "BIOMD0000000231_url.xml"
TEST_FILE10 = os.path.join(TEST_DIR, "test_file10.xml")
# test_file10: originally "BIOMD0000000253_url.xml"
TEST_FILE11 = os.path.join(TEST_DIR, "test_file11.xml")
# test_file12: originally "BIOMD0000000281_url.xml"
TEST_FILE12 = os.path.join(TEST_DIR, "test_file12.xml")
# test_file13: originally "BIOMD0000000035_url.xml"
TEST_FILE13 = os.path.join(TEST_DIR, "test_file13.xml")
# test_file_games_pp1: originally "BIOMD0000000383_url.xml"
TEST_FILE_GAMES_PP1 = os.path.join(TEST_DIR, "test_file_games_pp1.xml")
# test_file_games_pp2: originally "BIOMD0000000018_url.xml"
TEST_FILE_GAMES_PP2 = os.path.join(TEST_DIR, "test_file_games_pp2.xml")
# test_file_games_report1: originally "BIOMD0000000248_url.xml"
TEST_FILE_GAMESREPORT1 = os.path.join(TEST_DIR, "test_file_games_report1.xml")
# test_file_games_report2: originally "BIOMD0000000007_url.xml"
TEST_FILE_GAMESREPORT2 = os.path.join(TEST_DIR, "test_file_games_report2.xml")
# test_file_games_report3: originally "BIOMD0000000167_url.xml"
TEST_FILE_GAMESREPORT3 = os.path.join(TEST_DIR, "test_file_games_report3.xml")
NUM_REACTIONS = 111
NUM_PARAMETERS = 27
MAX_REACTANTS = 10
NUM_SPECIES = 32

############## FUNCTIONAL CONSTANTS ##############
# DataFrame Columns
VALUE = "value"

#
ARC_ARROW = "->"
EQUAL = "="
LABEL_SEPARATOR = ":"
LESSTHAN = "<"
KINETICS_SEPARATOR = ";"

# Reaction categories
REACTION_1_1 = "reaction_1_1"
REACTION_n_1 = "reaction_n_1"
REACTION_1_n = "reaction_1_n"
REACTION_n_n = "reaction_n_n"
REACTION_BOUNDARY = "reaction_boundary"
# for reduced reactions
REACTION_REDUNDANT = "reaction_redundant"
REACTION_ERROR = "reaction_error"
ReactionCategory = collections.namedtuple('ReactionCategory',
    'category predicate')
ReactionComponents = collections.namedtuple('ReactionComponents',
    'label reactants products')

# EmptySet in a reaction (ex. curated model 006)
EMPTYSET = "EmptySet"

# Reaction attributes for each arc in MESGraph
REACTION = "reaction"

# The selected category is the first one that first has
# a satisfied predicate
# Arguments:
# x: number of reactants, y: number of products
# z: sum of reactants stoichiometry
# w: sum of products stoichiometry
REACTION_CATEGORIES = [
    ReactionCategory(category=REACTION_1_1,
        predicate=lambda x,y,z,w: (x==1) and (y==1) and (z==w)),
    ReactionCategory(category=REACTION_1_n,
        predicate=lambda x,y,z,w: ((x==1) and (y>1) and (z==1.00)) \
                               or ((x==1) and (y==1) and (z<w))),
    ReactionCategory(category=REACTION_n_1,
        predicate=lambda x,y,z,w: ((x>1) and (y==1) and (w==1.00)) \
                               or ((x==1) and (y==1) and (z>w))),
    ReactionCategory(category=REACTION_n_n,
        predicate=lambda x,y,z,w: ((x>1) and (y>1)) 
                               or ((x==1) and (y>1) and (z!=1.00)) \
                               or ((x>1) and (y==1) and (w!=1.00))),
    ReactionCategory(category=REACTION_BOUNDARY,
        predicate=lambda x,y,z,w: (x==0) or (y==0) or (z==0) or (w==0)),
    ]
# Similar to REACTION_CATEGORIES but have different arguments r, p 
# and two additional categories - reaction_redundant and reaction_error
# r: list of reactants stoichiometry
# p: list of products stoichiometry
REACTION_SUMMARY_CATEGORIES = [
    ReactionCategory(category=REACTION_REDUNDANT,
        predicate=lambda x,y,r,p: (x==0) and (y==0)),
    ReactionCategory(category=REACTION_ERROR,
        predicate=lambda x,y,r,p: ((x==0) and (y!=0)) \
                                  or ((x!=0) and (y==0))),
    ReactionCategory(category=REACTION_1_1,
        predicate=lambda x,y,r,p: (x==1) and (y==1) and (sum(r)==sum(p))),
    
    ReactionCategory(category=REACTION_1_n,
        predicate=lambda x,y,r,p: (x==1) and (sum([r[0]<=e for e in p])==len(p))),
                     
    ReactionCategory(category=REACTION_n_1,
        predicate=lambda x,y,r,p: (y==1) and (sum([p[0]<=e for e in r])==len(r))),
    ]


# summarized version of a reaction
ReactionSummary = collections.namedtuple('ReactionSummary', 
                  'label reactants products category')


# building block for each MESGraph path
PathComponents = collections.namedtuple('PathComponents',
                                        'node1 node2 reactions')

# used for creating a report
NULL_STR = ""

# Default values for configuration file
CFG_DEFAULTS = {}
CFG_DEFAULT_PATH = PROJECT_DIR
CFG_DEFAULT_PATH = os.path.join(PROJECT_DIR,
    "kinetics_validator_cfg.yml")
CFG_SECTIONS = []

# Symbolic
SYM_T = "t"
SYM_X = "x"



"""Representation of an SBML function defintion."""

import libsbml


class FunctionDefinition():

    def __init__(self, sbml_function_definition,
          name=None, fid=None, arguments=None, body=None):
        """
        Parameters
        ----------
        sbml_function_definition: libsbml.FunctionDefinition
        name: str
        fid: str
        arguments: list-str
        body: str
        """
        self.sbml_function_definition = sbml_function_definition
        if self.sbml_function_definition is not None:
            name = self.sbml_function_definition.getName()
            fid = self.sbml_function_definition.getId()
            arguments = [
                  self.sbml_function_definition.getArgument(n).getName()
                  for n in 
                  range(self.sbml_function_definition.getNumArguments())]
            body = libsbml.formulaToL3String(
                  self.sbml_function_definition.getBody())
        self.function_name = name
        self.id = fid
        self.argument_names = arguments
        self.body = body
  
    def __repr__(self):
        argument_call = ",".join(self.argument_names)
        call_str = "%s(%s)" % (self.id, argument_call)
        return "%s: %s" % (call_str, self.body)
  
    @classmethod
    def makeBuiltinFunctions(cls):
        """
        Creates a function definition for the SBML delay function.

        Returns
        -------
        list-FunctionDefinition
        """
        functions = []
        # Delay
        functions.append(cls(None, name="delay", fid="delay",
               arguments=["a_species", "num"], body = "a_species"))
        # Exponential
        functions.append(cls(None, name="exp", fid="exp",
               arguments=["num"], body = "2.71828182**num"))
        #
        return functions



""" Utilities for sympymath. ***

SUffix conventions
  *Arr - numpy array
  *Mat - sympy.Matrix N X N
  *Vec - sympy.Matrix N X 1
  *s - list

Notes
1. By default, symbols are added to locals() of the caller.
   This can be changed by using the dct optional keyword.
   This means that all user callable functions must capture
   the correct symbol dictionary and use this explicitly
   in internal calls.

"""

import collections
import copy
import inspect
import numpy as np
import sympy

SMALL_VALUE = 1e-8

def _getDct(dct, frame):
    """
    Gets the dictionary for the frame.

    Parameters
    ----------
    dct: dictionary to use if non-None
    frame: stack frame

    Returns
    -------
    dict
    """
    if dct is None:
        #dct = frame.f_back.f_locals
        dct = frame.f_back.f_globals
    return dct

def addSymbols(symbolStr, dct=None, real=True, negative=False):
    """
    Adds symbols to the dictionary.

    Parameters
    ----------
    symbolStr: str
    dct: dict
        default: globals() of caller
    kwarg: optional arguments for sympy.symbols
    """
    newDct = _getDct(dct, inspect.currentframe())
    _addSymbols(symbolStr, newDct, real=real, negative=negative)

def _addSymbols(symbolStr, dct, **kwargs):
    symbols = symbolStr.split(" ")
    for symbol in symbols:
        dct[symbol] = sympy.Symbol(symbol, **kwargs)

def removeSymbols(symbolStr, dct=None):
    """
    Removes symbols from the dictionary.

    Parameters
    ----------
    symbolStr: str
    dct: dict
        Namespace dictionary
    """
    newDct = _getDct(dct, inspect.currentframe())
    _removeSymbols(symbolStr, newDct)

def _removeSymbols(symbolStr, dct):
    symbols = symbolStr.split(" ")
    for symbol in symbols:
        if symbol in dct.keys():
            del dct[symbol]

def isSymbol(obj):
    if "is_symbol" in dir(obj):
        return obj.is_symbol
    else:
        return False

def substitute(expression, subs=None):
    """
    Substitutes into the expression.

    Parameters
    ----------
    expression: sympy.expression
    subs: dict
       key: sympy.symbol
       value: number

    Returns
    -------
    sympy.expression
    """
    if subs is None:
        subs = {}
    if isNumber(expression):
        return expression
    if isSymbol(expression):
        if expression.name in subs:
            return subs[expression.name]
        elif expression in subs:
            return subs[expression]
        else:
            return expression
    expr = expression.copy()
    # Must be an expression
    symbolDct = {s.name: s for s in expression.free_symbols}
    # Update entry in substitution to be the same as the expression
    newSubs = dict(subs)
    for key, value in subs.items():
        if key.name in symbolDct.keys():
            del newSubs[key]
            newSubs[symbolDct[key.name]] = value
    expr = expr.subs(newSubs)
    return sympy.simplify(expr)

def evaluate(expression, isNumpy=True, **kwargs):
    """
    Evaluates the solution for the substitutions provided.

    Parameters
    ----------
    expression: sympy.Add/number
    isNumpy: bool
        return float or ndarray of float
    kwargs: dict
        keyword arguments for substitute

    Returns
    -------
    float/np.ndarray
    """
    if isSympy(expression):
        val = _evaluate(expression, isNumpy=isNumpy, **kwargs)
    else:
        val = expression
    return val

def _evaluate(expression, isNumpy=True, **kwargs):
    """
    Evaluates the solution for the substitutions provided.

    Parameters
    ----------
    expression: sympy.Add
    isNumpy: bool
        return float or ndarray of float
    kwargs: dict
        keyword arguments for substitute

    Returns
    -------
    float/np.ndarray
    """
    if isNumber(expression):
        if isNumpy:
            return expressionToNumber(expression)
        else:
            return expression
    # Evaluate
    expr = substitute(expression, **kwargs)
    # Symbol substitution can create a number
    if isNumber(expr):
        return expr
    val = expr.evalf()
    if hasSymbols(val):
        return val
    if isNumpy:
        if "rows" in dir(expression):
            result = np.array(val)
        else:
            try:
                result = float(val)
            except TypeError:
                result = complex(val)
    else:
        result = val
    return result

def mkVector(nameRoot, numRow, dct=None):
    """
    Constructs a vector of symbols.

    Parameters
    ----------
    nameRoot: str
        root name for elements of the vector
    numRow: int
    dct: dict
        Namespace dictionary

    Returns
    -------
    sympy.Matrix numRow X 1
    """
    newDct = _getDct(dct, inspect.currentframe())
    return _mkVector(nameRoot, numRow, newDct)

def _mkVector(nameRoot, numRow, dct):
    # Create the solution vector. The resulting vector is in the global name space.
    symbols = ["%s_%d" % (nameRoot, n) for n in range(numRow)]
    symbolStr = " ".join(symbols)
    addSymbols(symbolStr, dct=dct)
    return sympy.Matrix([ [s] for s in symbols])

def flatten(vec):
    """
    Converts a sympy N X 1 matrix to a list.

    Parameters
    ----------
    vec: symbpy.Matrix N X 1

    Returns
    -------
    list
    """
    return [ [v for v in z] for z in vec][0]

def vectorRoundToZero(vec):
    if vec.cols > 1:
        RuntimeError("Can only handle vectors.")
    newValues = [roundToZero(v) for v in vec]
    return sympy.Matrix(newValues)

def roundToZero(v):
    if isSympy(v):
        if not v.is_Number:
            return v
    if np.abs(v) < SMALL_VALUE:
        return 0
    return v

def solveLinearSystem(aMat, bMat):
    """
    Finds a solution to A*x = b. Chooses an arbitrary
    solution if multiple solutions exist.

    Parameters
    ----------
    aMat: sympy.Matrix (N X N)
        A
    bMat: sympy.Matrix (N X 1)
        b

    Returns
    -------
    sympy.Matrix: N X 1
    """
    numRow = aMat.rows
    dummyVec = mkVector("x", numRow)
    dummySymbols = [v for v in dummyVec]
    #
    system = aMat, bMat
    result = sympy.linsolve(system, *dummyVec)
    lst = flatten(result)
    # Handle case of multiple solutions
    subs = {s: 1 for s in lst if s in dummySymbols}
    return evaluate(sympy.Matrix(lst), subs=subs)

def expressionToNumber(expression):
    """
    Converts an exprssion to a numpy number.
    Throws an exception if it cannot be done.

    Parameters
    ----------
    expression: sympy.Add

    Returns
    -------
    float/complex

    Raises
    -------
    TypeError if not convertable to a number
    """
    # Convert expression to a number
    if isSympy(expression):
        val = expression.evalf()
        if val.is_real:
            val = float(val)
        else:
            val = complex(val)
    else:
        val = expression  # already a number
    # Eliminate small values
    if np.abs(val) < SMALL_VALUE:
        val = 0
    if np.abs(np.angle(val)) < SMALL_VALUE:
        val = np.sign(val) * np.abs(val)
    return val

def asRealImag(val):
    if isSympy(val):
        return val.as_real_imag()
    cmplxVal = complex(val)
    return (cmplxVal.real, cmplxVal.imag)

def isConjugate(val1, val2):
    """
    Tests for complex conjugates.

    Parameters
    ----------
    val1: number or expression
    val2: number or expression

    Returns
    -------
    bool
    """
    realImag = [asRealImag(val1), asRealImag(val2)]
    isSameReal = realImag[0][0] == realImag[1][0]
    isSameImag = realImag[0][1] == -realImag[1][1]
    return isSameReal and isSameImag

def _hasSymbols(val):
    if isSympy(val):
        return len(val.free_symbols) > 0
    else:
        return False

def hasSymbols(val):
    if isIndexable(val):
        trues = [_hasSymbols(v) for v in val]
    else:
        trues = [_hasSymbols(val)]
    return any(trues)

def isSympy(val):
    """
    Tests if this is a sympy object.

    Parameters
    ----------
    val: object

    Returns
    -------
    bool
    """
    properties = dir(val)
    return ("is_symbol" in properties) or ("evalf" in properties)

def isNumber(val):
    """
    Tests if this is a number in base python or sympy.

    Parameters
    ----------
    val: float/int/complex/sympy.expression

    Returns
    -------
    bool
    """
    try:
        _ = complex(val)
        return True
    except TypeError:
        return False

def isReal(val):
    if isSympy(val):
        return val.is_real
    return np.isreal(val)

def isComplex(val):
    return isNumber(val) and (not isReal(val))

def isZero(val):
    """
    Tests if a scalar, number or symbol, is 0.

    Parameters
    ----------
    val: number or symbol or expression

    Returns
    -------
    bool
    """
    if isSympy(val):
        try:
            val = expressionToNumber(val)
        except Exception:
            return False
    try:
        if np.isclose(np.abs(val), 0):
            return True
    except TypeError:
        newVal = complex(val)
        return np.abs(newVal) == 0

def isVecZero(vec):
    """
    Tests if a vector, numbers or symbols, are equal.

    Parameters
    ----------
    vec: sympy.Matrix (N X 1)

    Returns
    -------
    bool
    """
    trues = [isZero(e) for e in vec]
    return all(trues)

def solveLinearSingular(aMat, bVec, isParameterized=False, defaultValue=1):
    """
    Finds a solution to a linear system where the linear
    matrix may be singular.
    Parameter values are set to 1 if not isParameterized.

    Parameters
    ----------
    aMat: sympy.Matrix N X N
    bVec: sympy.Matrix N X 1
    isParameterized: bool
          Return the parameterized result

    Returns
    -------
    sympy.Matrix N X 1
    """
    solution = aMat.gauss_jordan_solve(bVec)
    solutionVec = solution[0]
    if not isParameterized:
        parameterMat = solution[1]
        for parameter in parameterMat:
            solutionVec = solutionVec.subs(parameter, defaultValue)
    solutionVec = solutionVec.evalf()
    return solutionVec

def isIndexable(obj):
    return "__getitem__" in dir(obj)

def recursiveEvaluate(obj, **kwargs):
    """
    Recursively evaluates symbols encountered in muteable indexables of type:
    list, np.array, sympy.Matrix (and maybe others)

    Parameters
    ----------
    obj: sympy.expr/number/indexable
        an indexable support indexing.
    kwargs: dict
        keyword arguments for evaluate

    Returns
    -------
    object with same structure as original
    """
    if isIndexable(obj):
        # Is a container of other objects
        if "tuple" in str(obj.__class__):
            newObj = list(obj)
        else:
            newObj = copy.deepcopy(obj)
        newerObj = copy.deepcopy(newObj)
        for idx, entry in enumerate(newObj):
            newerObj[idx] = recursiveEvaluate(entry, **kwargs)
        return newerObj
    # Do the numeric evaluation
    return evaluate(obj, **kwargs)

def recursiveEquals(obj1, obj2, **kwargs):
    """
    Recursively evaluates if two indexable, mutable objects are equal
    under subtituions.

    Parameters
    ----------
    obj: sympy.expr/number/indexable
        an indexable support indexing.
    kwargs: dict
        keyword arguments for evaluate

    Returns
    -------
    bool
    """
    if isIndexable(obj1) != isIndexable(obj2):
        return False
    if isIndexable(obj1):
        for entry1, entry2 in zip(obj1, obj2):
            if not recursiveEquals(entry1, entry2, **kwargs):
                return False
        return True
    # Do the numeric evaluation
    num1 = expressionToNumber(evaluate(obj1, **kwargs))
    num2 = expressionToNumber(evaluate(obj2, **kwargs))
    return np.isclose(num1, num2)

def vectorAsRealImag(vec):
    """
    Expresses a vector as the sum of real and imaginary parts.

    Parameters
    ----------
    vec - sympy.Matrix

    Returns
    -------
    sympy.Matrix, sympy.Matrix
        real vector, imaginary vector
    """
    reals = []
    imags = []
    for entry in vec:
        real, imag = asRealImag(entry)
        reals.append(real)
        imags.append(imag)
    return sympy.Matrix(reals),  sympy.Matrix(imags)

def eigenvects(mat):
    """
    Computes eigenvector results, handling near zero determinants.

    Parameters
    ----------
    mat: sympy.Matrix
    
    Returns
    -------
    eigenvect entries returned by sympy.eigenvects()
    """
    # Check if symbols are present
    if hasSymbols(mat):
        return mat.eigenvects()
    # Purely numeric matrix
    newMat = recursiveEvaluate(mat.as_mutable())
    return newMat.eigenvects()

def countNodes(epr):
        """
        Counts the number of nodes in the expression tree.

        Parameters
        ----------
        epr: sympy.expression
        
        Returns
        -------
        int
        """
        result = 1
        argLst = epr.args
        for arg in argLst:
            result += countNodes(arg)
        return result

def countSymbols(epr, symbols):
        """
        Counts the symbols that are present in the expression.

        Parameters
        ----------
        epr: expression
        symbols: list-Symbol
        
        Returns
        -------
        int
        """
        freeSymbols = epr.free_symbols
        return len(set(freeSymbols).intersection(symbols))

def countDctSymbols(dct, excludes=None):
        """
        Counts the symbols present in each expression in the dictionary.
 
        Parameters
        ----------
        dct: dict
            key: Symbol
            value: Expression
        excludes: list-symbol
        
        Returns
        -------
        dict
            key: Symbol
            value: int
        """
        symbols = list(dct.keys())
        if excludes is not None:
            symbols = list(set(symbols).difference(excludes))
        return {s: countSymbols(dct[s], symbols) for s in symbols}

def getDctSymbols(dct, symbols=None, excludes=None):
        """
        Lists the free symbols present in each expression in the dictionary.
 
        Parameters
        ----------
        dct: dict
            key: Symbol
            value: Expression
        excludes: list-symbol
        symbols: list-Symbol
        
        Returns
        -------
        dict
            key: Symbol
            value: int
        """
        if symbols is None:
            symbols = list(dct.keys())
        if excludes is not None:
            symbols = list(set(symbols).difference(excludes))
        else:
            excludes = []
        #
        result = {}
        for sym, epr in dct.items():
            if sym in excludes:
                continue
            lst = list(set(symbols).intersection(epr.free_symbols))
            result[sym] = sorted(lst, key=lambda s: s.name)
        #
        return result

def getDistinctSymbols(eqnDct, symbols=None, excludes=None):
    """
    Finds the distinct symbols on the expressions.

    Parameters
    ----------
    eqnDct: dict
    symbols: list-Symbol
        Defaults to eqnDct.keys()
    excludes: list-Symbol
    
    Returns
    -------
    list-symbol
    """
    if excludes is None:
        excludes = []
    dct = getDctSymbols(eqnDct, symbols)
    rhsSymbols = []
    [rhsSymbols.extend(v) for v in dct.values()]
    newRhsSymbols = [s for s in rhsSymbols if not s in excludes]
    lst = list(set(newRhsSymbols))
    lst = sorted(lst, key=lambda s: s.name)
    return lst

def getSymbols(epr):
    """
    Recursively extracts symbols from an expression.

    Parameters
    ----------
    epr: expression
    
    Returns
    -------
    list-Symbol
    """
    result = []
    for ele in epr.args:
        if ele.is_symbol:
            result.append(ele)
        else:
            result.extend(getSymbols(ele))
    return list(set(result))
     

def mkQuadraticRelaxation(systemDct):
    """
    Relaxes a set of system equations to linear system for quadratic system,
    a system consisting of terms of the form k*S or k*S*T or k*S**2.

    Parameters
    ----------
    systemDct
        key: Symbol
        value: expression
    
    Returns
    -------
    RelaxationResult
    """
    # Relaxation result
    # eqn: dict
    #     key: state variable
    #     value: expression
    # sub: dict
    #     key: new symbol
    #     value: expression to rever to original
    # mat: sympy.Matrix: matrix for linearized system (N X M)
    #     row: original state variables
    #     column: extended set of state variables
    #     value: linearized expression
    # vec: sympy.Matrix complete set of symbols in columns N X 1
    #     row: extended symbol
    RelaxationResult = collections.namedtuple("RelaxationResult",
          "eqn sub mat vec")
    # Add existing symbols
    freeSymbols = []
    [freeSymbols.extend(e.free_symbols) for e in systemDct.values()]
    [addSymbols(f.name, dct=globals()) for f in freeSymbols]
    # Add quadratic symbols
    systemStr = " ".join([str(e) for  e in systemDct.values()])
    stateSymbols = list(systemDct.keys())
    subs = {}  # key: new symbol; value: quadratic expression
    for sym1 in stateSymbols:
        for sym2 in stateSymbols:
            if sym1.name <= sym2.name:
                name = "%s%s" % (sym1, sym2)
                quadraticEpr = sym1 * sym2
                quadraticEprStr = str(quadraticEpr)
                if quadraticEprStr in systemStr:               
                    addSymbols(name, dct=globals())
                    subs[globals()[name]] = quadraticEpr
    subsStr = {str(k): str(v) for k, v in subs.items()}
    # Construct the vector of extended symbols
    extendedStateSyms = list(stateSymbols)
    extendedStateSyms.extend(list(subs.keys()))
    extendedStateVec = sympy.Matrix(extendedStateSyms)
    # Replace quadratic terms
    eprStrDct = {}
    linearizedDct = {}
    for sym, epr in systemDct.items():
        newEprStr = str(epr)
        for k, v in subsStr.items():
            newEprStr = newEprStr.replace(v, k)
        linearizedDct[sym] = eval(newEprStr)
    # Calculate Jacobian w.r.t. newSymbols to get matris
    stateEprVec = sympy.Matrix(list(linearizedDct.values()))
    mat = stateEprVec.jacobian(extendedStateVec)
    #
    return RelaxationResult(
          eqn=linearizedDct,
          sub=subs,
          mat=mat,
          vec=extendedStateVec,
          )
    return linearizedDct, subs

def isIterable(obj):
    try:
        [_ for _ in obj]
        return True
    except TypeError:
        return False

def findRecursive(epr, sym):
    """
    Recursively looks for symbol in epression.

    Parameters
    ----------
    epr: epression
    sym: symbol
    
    Returns
    bool
    -------
    """
    # Handle iterable
    if isIterable(epr):
        return any([findRecursive(a, sym) for a in epr])
    # Handle an expression
    if "args" in dir(epr):
        if str(epr) == str(sym):
            return True
        if sym in epr.args:
            return True
        return any([findRecursive(a, sym) for a in epr.args])
    # Handle other cases
    return str(epr) == str(sym)



"""
Symbolic representation of an SBML model

Provides:
* symbolic state equations: system
   index: symbol
   value: expression for derivative
* symbolic jacobian of state equations
   index: symbol1
   column: symbol2
   value: expression d symbol1/d symbol2

Key methods:
* substatitue method provides for substituting symbols with roadrunner values
  (e.g., values of parameters in roadrunner or values of species in roadrunner)
  or user designated values.
* get provides the symbol and value information for a name (ID in libsbml)
* set changes the value information for a name (ID in libsbml)

Key properties:
* system is a Series of the system equations. The index is the variable,
  and the value is the expression.
* jacobian is a DataFrame for the Jacobian. Indexes are the variable whose
  expression is being differentiated and columns are the variables in the
  denominator of the differential
* roadrunner is the RoadRunner instance. Running simulations changes
  the values of variables that are substituted.
* namespace_dct is a dictionary with the namespace in which roadrunner
  variables are defined as symbols.
"""

"""
TO DO:
0. Update README with use case.
.1 Get Jin to change a file so she has commit history
1. Create package
"""

from symmathSBML.symmath_base import SymmathBase
from symmathSBML import msgs
from symmathSBML.persister import Persister

import collections
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import sympy
import seaborn as sns
import tellurium as te


ANT = "ant"
XML = "xml"
TIME = "time"


TYPE_MODEL = "type_model"  # libsbml model
TYPE_XML = "type_xml"  # XML string
TYPE_ANTIMONY = "type_xml"  # Antimony string
TYPE_FILE = "type_file" # File reference

# filename: name of file processed
# number: index of item
# model: libsbml.Model
IteratorItem = collections.namedtuple('IteratorItem',
    'filename number model')
# nme: str name
# sym: sympy.Symbol
# rr: value in roadrunner
SymbolInfo = collections.namedtuple("SymbolInfo", "nme sym rr")


class SymmathSBML(SymmathBase):

    def __init__(self, model_reference, is_suppress_warnings=False):
        """
        Parameters
        ----------
        model_reference: str/roadrunner
            str: path to ant file, path to XML file, url to xml file, ant str
        is_suppress_warnings: bool
        """
        def mkAntimonyStr():
            try:
                self.roadrunner = te.loada(model_reference)
                xml_model_reference = self.roadrunner.getSBML()
                return xml_model_reference
            except:
                raise ValueError("Invalid model reference")
        ##### PUBLIC #####
        super(SymmathSBML, self).__init__(model_reference,
              is_suppress_warnings=is_suppress_warnings)
        # stoichiometry_mat - stoichiometry matrix including fixed species
        # kinetics - pd.Series of kinetics expressions
        # system - Series of the system expressions, indexed by species
        # jacobian - Dataframe of the jacobian expressions
        self.stoichiometry_mat, self.kinetics, self.system, self.jacobian =  \
              self.mkSymbolSystem()

    def copy(self):
        """
        Creates a copy of the object.
        
        Returns
        -------
        symmathSBML
        """
        new_symmath = SymmathSBML(self.antimony)
        new_symmath.model_reference = self.model_reference
        # Adjust the simulation time
        new_symmath.roadrunner.reset()
        _ = new_symmath.roadrunner.simulate(0, self.roadrunner.model.getTime())
        # Save the content of the roadrunner instance
        dct = self._getRoadrunnerDct()
        new_symmath.set(dct)
        #
        return new_symmath

    def serialize(self, path):
        """
        Serializes the current object to a file.

        Parameters
        ----------
        path: str
            Path to serialization file
        """
        persister = Persister(path)
        time = self.roadrunner.model.getTime()
        roadrunner_dct = self._getRoadrunnerDct()
        items = [self.antimony, roadrunner_dct, time]
        persister.set(items)

    @classmethod
    def deserialize(cls, path):
        """
        Creates a SymmathSBML object from a file serialization.

        Parameters
        ----------
        path: str
            Path to serialization file
        
        Returns
        -------
        SymmathSBML
        """
        persister = Persister(path)
        [antimony, roadrunner_dct, time] = persister.get()
        symmath = cls(antimony)
        symmath.roadrunner.reset()
        _ = symmath.roadrunner.simulate(0, time)
        symmath.set(roadrunner_dct)
        return symmath

    def equals(self, other):
        """
        Checks that they have the same information

        Parameters
        ----------
        other: SymmathSBML
        
        Returns
        -------
        bool
        """
        val_bool = self.antimony == other.antimony
        val_bool = val_bool and all([s1.id == s2.id for s1, s2 in
              zip(self.species, other.species)])
        val_bool = val_bool and all([s1 == s2 for s1, s2 in
              zip(self.species_names, other.species_names)])
        val_bool = val_bool and all([p1.id == p2.id for p1, p2 in
              zip(self.parameters, other.parameters)])
        val_bool = val_bool and all([p1 == p2 for p1, p2 in
              zip(self.parameter_names, other.parameter_names)])
        val_bool = val_bool and all([r1.id == r2.id for r1, r2 in
              zip(self.reactions, other.reactions)])
        val_bool = val_bool and self.system.equals(other.system)
        val_bool = val_bool and self.jacobian.equals(other.jacobian)
        # Compare roadrunner values
        val_bool = val_bool and np.isclose(
               self.roadrunner.model.getTime(), other.roadrunner.model.getTime())
        dct = self.getParameterNameDct()
        dct.update(self.getSpeciesNameDct())
        for name, value in dct.items():
            val_bool = val_bool and np.isclose(other.roadrunner[name], value)
        return val_bool

    @staticmethod
    def _convert(value, dct):
        """
        Converts a value or symbol or expression with substitution.
   
        Parameters
        ----------
        dct: dict
            key: Symbol
            value: object
        
        Returns
        -------
        expression or number
        """
        if "is_Symbol" in dir(value):
            new_value = value.subs(dct).simplify()
            if len(new_value.free_symbols) == 0:
                new_value = float(new_value)
        else:
            new_value = float(value)
        return new_value

    def _getSubstitutionDct(self, is_sub_parameters=True,
          is_sub_species=False, is_sub_others=True):
        """
        Constructs a substitution dictionary.
   
        Parameters
        ----------
        
        Returns
        -------
        dict
        """
        dct = {}
        if is_sub_parameters:
            dct.update(self.getParameterSymbolDct())
        if is_sub_species:
            dct.update(self.getSpeciesSymbolDct())
        if is_sub_others:
            dct.update(self.getOtherSymbolDct())
        return dct
     
    def getJacobian(self, is_sub_parameters=True,
          is_sub_species=False, is_sub_others=True):
        """
        Substitutes into the Jacobian as directed.
   
        Parameters
        ----------
        is_sub_parameters: bool
            substitute all parameters
        is_sub_species: bool
            substitute all species
        is_sub_others: bool
            substitute the non-species, non-parameters
        """
        dct = self._getSubstitutionDct(
              is_sub_parameters=is_sub_parameters,
              is_sub_species=is_sub_species,
              is_sub_others=is_sub_others)
        jacobian = self.jacobian.applymap(
              lambda v: sympy.simplify(self._convert(v, dct)))
        return jacobian
     
    def getSystem(self, is_sub_parameters=True,
          is_sub_species=False, is_sub_others=True):
        """
        Substitutes into the System equations.
   
        Parameters
        ----------
        is_sub_parameters: bool
            substitute all parameters
        is_sub_species: bool
            substitute all species
        is_sub_others: bool
            substitute the non-species, non-parameters
        """
        dct = self._getSubstitutionDct(
              is_sub_parameters=is_sub_parameters,
              is_sub_species=is_sub_species,
              is_sub_others=is_sub_others)
        return self.system.apply(lambda v: self._convert(v, dct))
     
    def getKinetics(self, is_sub_parameters=True,
          is_sub_species=False, is_sub_others=True):
        """
        Substitutes into the System kinetics expressions
        for each reaction.
   
        Parameters
        ----------
        is_sub_parameters: bool
            substitute all parameters
        is_sub_species: bool
            substitute all species
        is_sub_others: bool
            substitute the non-species, non-parameters
        """
        dct = self._getSubstitutionDct(
              is_sub_parameters=is_sub_parameters,
              is_sub_species=is_sub_species,
              is_sub_others=is_sub_others)
        return self.kinetics.apply(lambda v: self._convert(v, dct))

    def mkSymbolSystem(self):
        """
        Creates an ODEModel.
    
        Returns
        -------
        np.array, pd.Series, pd.Series, pd.DataFrame
            stoichiometry_mat, kinetics, system, jacobian
        """
        def evaluate(reaction):
            return eval(reaction.kinetic_law.expanded_formula, namespace_dct)
        # Create sympy expressions for kinetic laws
        reaction_epr_dct = {}
        # Add sympy symbols
        self.namespace_dct["log"] = sympy.log
        self.namespace_dct["sqrt"] = sympy.sqrt
        # Process reactions
        namespace_dct = dict(self.namespace_dct)  # Don't modify namespace
        for reaction in self.reactions:
            key = namespace_dct[reaction.id]
            for _ in range(10):  # Number of attempts to resolve names
                is_done = False
                value = None
                try:
                    value = evaluate(reaction)
                    is_done = True
                except NameError as err:
                    parts = str(err).split(" ")
                    name = parts[1].replace("'", "")
                    namespace_dct[name] = sympy.Symbol(name)
                    # Keept this in the namespace
                    self.namespace_dct[name] = namespace_dct[name]
                    msg = "Name %s undefined in reaction %s. Set to 0."  \
                          % (name, str(reaction))
                    msgs.warn(msg,
                          is_suppress_warnings=self.is_suppress_warnings)
                if is_done:
                    break
            reaction_epr_dct[key] = value
        # Express species fluxes in terms of reaction fluxes
        reaction_vec = sympy.Matrix(list(reaction_epr_dct.keys()))
        # Ensure we get all boundary species in the stoichiometrymatrix
        # FIXME: Does this include assignment rules?
        boundary_species = self.roadrunner.getBoundarySpeciesIds()
        for name in boundary_species:
            self.roadrunner.setBoundary(name, False)
        try:
            stoichiometry_mat = self.roadrunner.getFullStoichiometryMatrix()
        except Exception:
            msgs.warn("Stoichiometery matrix does not exist.",
                  is_suppress_warnings=self.is_suppress_warnings)
            stoichiometry_mat = None
        # Only rate rules
        if stoichiometry_mat is None:
            reaction_epr_dct = {}
            system_dct = {}
            self._updateSystemDctWithRules(system_dct)
        # Has reactions
        else:    
            species_reaction_vec = stoichiometry_mat * reaction_vec
            species_in_reactions = self.getSpeciesInReactions()
            rownames = list(stoichiometry_mat.rownames)
            species_in_reactions = sorted(species_in_reactions,
                  key=lambda n: rownames.index(n))
            species_reaction_dct = {self.namespace_dct[s]: 
                  e for s, e in zip(species_in_reactions, species_reaction_vec)}
            # Construct the system equations
            system_dct = {s: sympy.simplify(
                  species_reaction_dct[s].subs(reaction_epr_dct))
                for s in species_reaction_dct.keys()}
            # Handle rules
            self._updateSystemDctWithRules(system_dct)
        system = pd.Series(system_dct.values(), index=system_dct.keys())
        # Construct the Jacobian
        variables = list(system.index)
        jac_dct = {v: [] for v in variables}
        for var1 in variables:
            epr = system.loc[var1]
            for var2 in variables:
                jac_dct[var2].append(sympy.diff(epr, var2))
        jacobian = pd.DataFrame(jac_dct, index=variables)
        #
        kinetics = pd.Series(reaction_epr_dct, dtype=object)
        return stoichiometry_mat, kinetics, system, jacobian

    def _updateSystemDctWithRules(self, system_dct):
        """
        Updates the system dictionary (and self.namespace_dct) as a result
        of assignment rules and rate rules.

        Parameters
        ----------
        system_dct: dict
            key: symbol
            value: expression
        """
        for num in range(self.model.getNumRules()):
            rule = self.model.getRule(num)
            formula_str = rule.getFormula()
            try:
                expression = eval(formula_str, self.namespace_dct)
            except Exception as excp:
                import pdb; pdb.set_trace()
            variable = rule.getVariable()
            self.species_names.append(variable)
            if not variable in self.namespace_dct.keys():
                self.namespace_dct[variable] = sympy.Symbol(variable)
            symbol = self.namespace_dct[variable]
            if "Rate" in str(type(rule)):
                system_dct[symbol] = expression
            elif "Assignment" in str(type(rule)):
                self.namespace_dct[variable] = expression
            else:
                raise RuntimeError("Unknown rule type")

    def setTime(self, time):
        """
        Sets the time for the simulation by running the simulation to that time.

        Parameters
        ----------
        time: float
        """
        self.roadrunner.reset()
        self.roadrunner.simulate(0, time)

    def calculateJacobianSensitivity(self, is_normalized=True):
        """
        Calculates the sensitivity of the species (state) variable w.r.t. all
        other species by summing derivatives and evaluating at the current
        value of the state vector.

        Parameters
        ----------
        is_normalized: bool
            divide by the value
        
        Returns
        -------
        DataFrame
            columns: species column in Jacobian
            index: species used in derivative denominator
            values: sympy expression
        """
        # Initializations
        symbols = list(self.jacobian.index)
        dct = {s: [] for s in symbols}
        # Construct substitution dictionary
        subs_dct = self.getParameterSymbolDct()
        subs_dct.update(self.getSpeciesSymbolDct())
        # Take derivatives
        for s1 in symbols: # Row
            s1_arr = np.array(self.jacobian[s1])
            row_epr = s1_arr.dot(self.getSpeciesArray())
            row_value = row_epr.subs(subs_dct)
            if row_value == 0:
                row_value = 10e5
            for s2 in symbols:  # Columns
                deriv = sympy.diff(row_epr, s2).subs(subs_dct)
                if is_normalized:
                    deriv = deriv/row_value
                if deriv.is_number:
                    deriv = float(deriv)
                dct[s1].append(deriv)
        # Return the result
        df = pd.DataFrame(dct, index=symbols, columns=symbols)
        return df
            
    def getSpeciesArray(self):
        return np.array([self.get(s).rr for s in self.species_names])

    def plotJacobianSensitivityHeatmap(self, title="", is_plot=True,
          vmin=None, vmax=None):
        """
        Plots a heatmap of the derivative of each row in the Jacobain
        multiplied by the current value of the state vector.

        Parameters
        ----------
        title: str
        """
        df = self.calculateJacobianSensitivity()
        max_val = max([np.abs(v) for v in df.values.flatten()])
        if vmin is None:
            vmin = -max_val
        if vmax is None:
            vmax = max_val
        columns = list(df.columns)
        columns.reverse()
        df_new = df.iloc[::-1]
        sns.heatmap(df_new, cmap="seismic", vmin=vmin, vmax=vmax)
        plt.xlabel("State Evaluated")
        plt.ylabel("Derivative w.r.t. State")
        plt.title(title)
        if is_plot:
            plt.show()

    def addNamespace(self, dct):
        """
        Adds the symbol namespace to the dictionary.

        Parameters
        ----------
        dct: dict       
        """
        dct.update(self.namespace_dct)

    def removeNamespace(self, dct):
        """
        Removes the symbol namespace to the dictionary.

        Parameters
        ----------
        dct: dict       
        """
        for key in self.namespace_dct:
            if key in dct.keys():
                del dct[key]

    def getSpeciesInReactions(self):
        """
        Get names of all species that appear as reactants or products.
        
        Returns
        -------
        list-str
        """
        names = []
        for reaction in self.reactions:
            names.extend([str(s.getSpecies()) for s in reaction.reactants])
            names.extend([str(s.getSpecies()) for s in reaction.products])
        names = list(set(names))
        names.sort()
        return names



"""Creates a Roadrunner instance from a model reference."""

import tellurium as te

ANT = "ant"
XML = "xml"

def makeRoadrunner(model_reference):
    """
    Creates a roadrunner instance from a model reference.

    Parameters
    ----------
    model_reference: str/ExtendedRoadrunner
        Roadrunner object
        URL (http:...)
        Antimony file (extension is .ant)
        XML file (extension is .xml)
        XML string
        Antimony string
    
    Returns
    -------
    ExtendedRoadrunner object
    """
    if "RoadRunner" in str(type(model_reference)):
        return model_reference
    #
    if not isinstance(model_reference, str):
        Raise ("Invalid model reference")
    #
    if model_reference[0:4] == "http":
        return te.loadSBMLModel(model_reference)
    #
    parts = model_reference.split(".")
    if len(parts) == 2:
        if parts[1] == XML:
            return te.loadSBMLModel(model_reference)
        elif parts[1] == ANT:
            return te.loadAntimonyModel(model_reference)
        else:
            # Assume string for antimony model
            return te.loada(model_reference)
    else:
        if XML in model_reference[0:10]:
            return te.loads(model_reference)
        return te.loada(model_reference)



'''Persists an object as a Pickle file.'''

import pickle
import os


class Persister(object):
  """
  A light wrapper around pickle to persist an object.
  """

  def __init__(self, path):
    """
    :param str path: path for the pickle file persistence.
    """
    self.path = path

  def __repr__(self):
    return "Persister for %s" % self.path

  def isExist(self):
    return os.path.isfile(self.path)
    
  def set(self, an_object):
    with open(self.path, 'wb') as fd:
      pickle.dump(an_object, fd)

  def get(self):
    """
    :return object:
    Will throw FileNotFoundError if self.path does not exist
    """
    with open(self.path, 'rb') as fd:
      return pickle.load(fd)

  def remove(self):
    """
    Deletes the persistence file.
    """
    if self.isExist():
      os.remove(self.path)



"""Utitility functions"""

from symmathSBML import constants as cn
from symmathSBML import msgs
from symmathSBML.symmath_sbml import SymmathSBML
from symmathSBML.symmath_base import IteratorItem

import os.path
import zipfile


def getZipfilePaths(data_dir=cn.DATA_DIR,
        zip_filename=cn.BIOMODELS_ZIP_FILENAME):
      """
      :param str data_dir: absolute path of the directory containing
          the xml files
      :param str zip_filename: name of the zipfile to process
      :return list-str, ZipFile: list of file paths, ZipFile object for file
      """
      path = os.path.join(data_dir, zip_filename)
      zipper = zipfile.ZipFile(path, "r")
      files = [f.filename for f in zipper.filelist]
      return files, zipper

def modelIterator(initial=0, final=1000,
        data_dir=cn.DATA_DIR,
        zip_filename=cn.BIOMODELS_ZIP_FILENAME):
    """
    Iterates across all models in a data directory.
    :param int initial: initial file to process
    :param int final: final file to process
    :param str data_dir: absolute path of the
        directory containing
        the xml files
    :param str zip_filename: name of the zipfile to process.
        If None, then looks for XML files in the directory.
    :return IteratorItem:
    """
    files, zipper = getZipfilePaths(
          data_dir=data_dir, zip_filename=zip_filename)
    # Functions for file types
    def readXML(filename):
        path = os.path.join(data_dir, filename)
        with open(path, 'r') as fd:
            lines = ''.join(fd.readlines())
        return lines
    def readZip(filename):
        with zipper.open(filename, 'r') as fid:
            lines = fid.read()
        return lines
    #
    if zip_filename is not None:
        read_func = readZip
    else:
        read_func = readXML
    begin_num = max(initial, 0)
    num = begin_num - 1
    end_num = min(len(files), final)
    for filename in files[begin_num:end_num]:
        num += 1
        lines = read_func(filename)
        if isinstance(lines, bytes):
          lines = lines.decode("utf-8")
        try:
            model = SymmathSBML(lines)
        except NameError as err:
            msg = "%s: %s" % (filename, str(err))
            msgs.error(msg)
        iterator_item = IteratorItem(filename=filename,
            model=model, number=num)
        yield iterator_item



"""Commonly used utilities."""

import os

TYPE_ANTIMONY = "type_antimony"
TYPE_XML = "type_xml"
TYPE_FILENAME = "type_filename"
XML_HEADER = '<?xml version="1.0" encoding="UTF-8"?>'
DEFAULT_MSG = "An error occurred in an input file."

def getXML(model_reference):
      """
      :param str model_reference:
          the input may be a file reference or a model string
          or TextIOWrapper
              and the file may be an xml file or an antimony file.
          if it is a model string, it may be an xml string or antimony.
      :raises IOError: Error encountered reading the SBML document
      :return str SBML xml"
      """
      # Check for a file path
      model_str = ""
      if isinstance(model_reference, str):
        if os.path.isfile(model_reference):
          with open(model_reference, 'r') as fd:
            lines = fd.readlines()
          model_str = ''.join(lines)
      if len(model_str) == 0:
        if "readlines" in dir(model_reference):
          lines = model_reference.readlines()
          if isinstance(lines[0], bytes):
            lines = [l.decode("utf-8") for l in lines]
          model_str = ''.join(lines)
          model_reference.close()
        else:
          # Must be a string representation of a model
          model_str = model_reference
      # Process model_str into a model
      if not "<sbml" in model_str:
        # Antimony
        raise ValueError("Invalid SBML model.")
      return model_str

def checkSBMLDocument(document, model_reference=""):
      if document.getNumErrors() > 0:
        raise ValueError("Errors in SBML document\n%s"
            % model_reference)

def tokenize(text, tokens):
    """
    Creates a list of strings separated by tokens.

    Parameters
    ----------
    text: str
    tokens: list-char

    Returns
    -------
    list-str
    """
    result = []
    start_pos = 0
    end_pos = 0
    for pos, char in enumerate(text):
        if char in tokens:
            result.append(text[start_pos:end_pos])
            result.append(char)
            start_pos = pos + 1
        end_pos = pos + 1
    if start_pos < len(text):
        result.append(text[start_pos:end_pos])
    return result

def detokenize(tokenizeds):
    """
    Creates a string from the tokenized list.

    Parameters
    ----------
    list-str

    Returns
    -------
    str
    """
    return "".join(tokenizeds)



"""Generate messages for the application."""

import sys

def error(text):
    print("***Error. %s" % text)
    sys.exit()

def warn(text, is_suppress_warnings=False):
    if not is_suppress_warnings:
        print("***Warning. %s" % text)



from setuptools import setup, find_packages
import os.path
import codecs

with open("README.md", "r") as fh:
    long_description = fh.read()

# The following two methods were copied from
# https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

# parse_requirements() returns generator of pip.req.InstallRequirement objects
INSTALL_REQUIREMENTS = [
    "coverage",
    "control",
    "jupyterlab",
    "matplotlib",
    "nose",
    "numpy",
    "pandas",
    "pip",
    "pylint",
    "python-libsbml",
    "seaborn",
    "sympy",
    "tellurium",
    "tk",
    "urllib3",
]


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            parts = line.split(" ")
            version = parts[2]
            version = version.replace('"', '')
            version = version.replace("'", '')
            print ("Processing version %s" % version)
            return version
    else:
        raise RuntimeError("Unable to find version string.")

def doSetup(install_requires):
  try:
    setup(
        name='symmathSBML',
        version=get_version("symmathSBML/_version.py"),
        author='Joseph L. Hellerstein, Jin Liu',
        author_email='jlheller@uw.edu',
        packages=find_packages(exclude=['tests', 'sym', 'images', 'data',
            'notebook', 'docs']),
        url='https://github.com/ModelEngineering/symmathSBML',
        description='Symbolic mathematics for SBML Models.',
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        package_dir={'symmathSBML': 'symmathSBML'},
        python_requires='>=3.6',
        install_requires=install_requires,
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',      # Define that your audience are developers
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'License :: OSI Approved :: MIT License',   # Again, pick a license
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
          ],
        )
  except Exception as exp:
    import pdb; pdb.set_trace()
    pass


if __name__ == '__main__':
  doSetup(INSTALL_REQUIREMENTS)



"""Data initializations, getter, setter methods"""

from symmathSBML.reaction import Reaction
from symmathSBML.function_definition import FunctionDefinition
import symmathSBML.sympy_util as su
from symmathSBML import util
from symmathSBML.make_roadrunner import makeRoadrunner

import collections
import libsbml
import tellurium as te


ANT = "ant"
XML = "xml"
TIME = "time"


TYPE_MODEL = "type_model"  # libsbml model
TYPE_XML = "type_xml"  # XML string
TYPE_ANTIMONY = "type_xml"  # Antimony string
TYPE_FILE = "type_file" # File reference

# filename: name of file processed
# number: index of item
# model: libsbml.Model
IteratorItem = collections.namedtuple('IteratorItem',
    'filename number model')
# nme: str name
# sym: sympy.Symbol
# rr: value in roadrunner
SymbolInfo = collections.namedtuple("SymbolInfo", "nme sym rr")


class SymmathBase(object):

    def __init__(self, model_reference, is_suppress_warnings=False):
        """
        Initializes instance variables
        :param str model_reference: string or SBML file or Roadrunner object
        """
        ##### PUBLIC #####
        self.is_suppress_warnings = is_suppress_warnings
        self.model_reference = model_reference
        self.reactions = []  # Python wrapper for Reaction
        self.species = []  # libsbml Species
        self.roadrunner = None  # Roadrunner object
        self.namespace_dct = {}  # Namespae of symbols added
        self.roadrunner = makeRoadrunner(model_reference)
        # Construct the libsbml document
        xml_model_reference = self.roadrunner.getSBML()
        xml = util.getXML(xml_model_reference)
        reader = libsbml.SBMLReader()
        document = reader.readSBMLFromString(xml)
        util.checkSBMLDocument(document, model_reference=model_reference)
        self.model = document.getModel()
        # Other public
        self.antimony = self.roadrunner.getAntimony()
        self.species = [self.model.getSpecies(nn)
            for nn in range(self.model.getNumSpecies())]
        # This may also include the variable in a rate rule
        self.species_names = [s.id for s in self.species]
        self.parameters = [self.model.getParameter(nn)
            for nn in range(self.model.getNumParameters())]
        self.parameter_names = [p.id for p in self.parameters]
        self.function_definitions = self.getFunctionDefinitions()
        self.reactions = [Reaction(self.model.getReaction(nn),
            function_definitions=self.function_definitions)
            for nn in range(self.model.getNumReactions())]
        self._populateNamespace()

    def _getRoadrunnerDct(self):
        """
        Returns
        -------
        dict
            Dictionary of name, value pairs in Roadrunner
        """
        dct = self.getParameterNameDct()
        dct.update(self.getSpeciesNameDct())
        return dct

    def get(self, names):
        """
        Provides the SymbolInfo for the name.

        Parameters
        ----------
        name: str/list-str

        Returns
        -------
        SymbolInfo/list-SymbolInfo
        """
        if isinstance(names, str):
            return self._getSymbolInfo(names)
        symbol_infos = []
        for name in names:
            symbol_infos.append(self._getSymbolInfo(name))
        return symbol_infos

    def _getSymbolInfo(self, name):
        """
        Provides the SymbolInfo for the name. A 0 is returned
        if the name is not in self.namespace or self.roadrunner.

        Parameters
        ----------
        name: str

        Returns
        -------
        SymbolInfo
        """
        if not name in self.roadrunner.keys():
            rr_value = 0
        else:
            rr_value = self.roadrunner[name]
        if name in self.namespace_dct.keys():
            symbol = self.namespace_dct[name]
        else:
            symbol = None
        return SymbolInfo(nme=name, rr=rr_value, sym=symbol)

    def set(self, dct):
        """
        Sets the values of names and values.

        Parameters
        ----------
        dct: dict
            key: str (name)
            value: object (value)
        """
        for name, value in dct.items():
            if not name in self.roadrunner.keys():
                raise ValueError("%s is not known to this roadrunner instance."
                      % name)
            self.roadrunner[name] = value

    def getParameterSymbolDct(self):
        """
        Returns
        -------
        dict
            key: symbol (parameter)
            value: value in self.roadrunner
        """
        return {self.namespace_dct[p]:
              self.roadrunner[p] for p in self.parameter_names}

    def getParameterNameDct(self):
        """
        Returns
        -------
        dict
            key: str (parameter)
            value: value in self.roadrunner
        """
        return {p: self.roadrunner[p] for p in self.parameter_names}

    def getSpeciesSymbolDct(self):
        """
        Returns
        -------
        dict
            key: symbol (species)
            value: value in self.roadrunner
        """
        return {self.namespace_dct[s]:
              self.roadrunner[s] for s in self.species_names}

    def _getOtherNames(self):
        """
        Names of non-species, non-parameters known to roadrunner.
        
        Returns
        -------
        list-str
        """
        keys = set(self.namespace_dct.keys()).difference(self.species_names)
        # DOn't include internal names
        keys = [k for k in set(keys).difference(self.parameter_names)
              if str(k)[0:2] != "__"]
        return list(keys)

    def getOtherSymbolDct(self):
        """
        Dictionary of symbols for non-species, non-parameters.
   
        Returns
        -------
        dict
            key: symbol
            value: value in self.roadrunner
        """
        names = self._getOtherNames()
        return {self.namespace_dct[s]: self.roadrunner[s]
              if s in self.roadrunner.keys() else None for s in names}

    def getOtherNameDct(self):
        """
        Dictionary of symbols for non-species, non-parameters.
   
        Returns
        -------
        dict
            key: symbol
            value: value in self.roadrunner
        """
        names = self._getOtherNames()
        return {n: self.roadrunner[n]
              if n in self.roadrunner.keys() else None for n in names}

    def getSpeciesNameDct(self):
        """
        Returns
        -------
        dict
            key: str (species)
            value: value in self.roadrunner
        """
        return {s: self.roadrunner[s] for s in self.species_names}

    def getFunctionDefinitions(self):
        """
        Returns all function definitions in the model.py

        Returns
        -------
        list-libsbml.FunctionDefinition
        """
        sbml_definitions =[self.model.getFunctionDefinition(n) for n
            in range(self.model.getNumFunctionDefinitions())]
        function_definitions = [FunctionDefinition(s) for s in sbml_definitions]
        # Include the builtin functions
        function_definitions.extend(FunctionDefinition.makeBuiltinFunctions())
        return function_definitions

    def getReaction(self, an_id):
        """
        Finds a reaction with the specified id.
        :param str an_id: id for the reaction
        :return Reaction/None:
        """
        return self._getInstance(self.reactions, an_id)

    def getSpecies(self, an_id):
        """
        Finds and returns the species with given name
        :param str an_id:
        Return None if there is no such species.
        """
        return self._getInstance(self.species, an_id)

    def getParameter(self, an_id):
        """
        Finds and returns the Parameter with given name
        :param str an_id:
        Return None if there is no such parameter.
        """
        return self._getInstance(self.parameters, an_id)

    def _getInstance(self, a_list, an_id):
        """
        Finds and returns the species with given name
        Return None if there is no such molecules
        :param str id:
        """
        results = [e for e in a_list if e.getId() == an_id]
        if len(results) > 1:
          raise ValueError(
              "Two instances have the same id: %s" %
              id)
        if len(results) == 0:
          return None
        return results[0]

    def _populateNamespace(self):
        """
        Populates the namespace with sympy symbols
        from roadrunner species and parameters.
        """
        names = list(self.parameter_names)
        species_names = list(self.species_names)
        names.extend(species_names)
        for reaction in self.reactions:
            name = reaction.id
            if name[0] == "_":
                name = name[1:]
            reaction.id = name
        reaction_names = [r.id for r in self.reactions]
        names.extend(reaction_names)
        names.extend(self.roadrunner.getCompartmentIds())
        nameStr = " ".join(names)
        su.addSymbols(nameStr, dct=self.namespace_dct)
